"""Provides utility functions for use across Graphein."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import os
import tempfile
from functools import lru_cache, partial
from multiprocessing import Pool
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Type, Union
from urllib.error import HTTPError
from urllib.request import urlopen

import networkx as nx
import numpy as np
import pandas as pd
import requests
import wget
from biopandas.pdb import PandasPdb
from loguru import logger as log
from tqdm import tqdm

from .resi_atoms import BACKBONE_ATOMS, RESI_THREE_TO_1

pdb_df_columns = [
    "record_name",
    "atom_number",
    "blank_1",
    "atom_name",
    "alt_loc",
    "residue_name",
    "blank_2",
    "chain_id",
    "residue_number",
    "insertion",
    "blank_3",
    "x_coord",
    "y_coord",
    "z_coord",
    "occupancy",
    "b_factor",
    "blank_4",
    "segment_id",
    "element_symbol",
    "charge",
    "line_idx",
]


class ProteinGraphConfigurationError(Exception):
    """
    Exception when an invalid Graph configuration if provided to a downstream
    function or method.
    """

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return self.message


@lru_cache()
def get_obsolete_mapping() -> Dict[str, str]:
    """Returns a dictionary mapping obsolete PDB codes to their replacement.

    :return: Dictionary mapping obsolete PDB codes to their replacement.
    :rtype: Dictionary[str, str]
    """
    obs_dict: Dict[str, str] = {}

    response = urlopen(
        "https://files.wwpdb.org/pub/pdb/data/status/obsolete.dat"
    )
    for line in response:
        entry = line.split()
        if len(entry) == 4:
            obs_dict[entry[2].lower().decode("utf-8")] = (
                entry[3].lower().decode("utf-8")
            )
        elif len(entry) == 3:
            obs_dict[entry[2].lower().decode("utf-8")] = ""
    return obs_dict


def read_fasta(file_path: str) -> Dict[str, str]:
    """
    Reads a FASTA file and returns a dictionary mapping sequence names to
    their identifiers.

    :param file_path: Path to FASTA file.
    :type file_path: str
    :return: Dictionary mapping sequence names to their identifiers.
    :rtype: Dict[str, str]
    """
    sequences = {}
    current_sequence_name = None
    current_sequence = ""

    with open(file_path, "r") as file:
        for line in file:
            if line.startswith(">"):
                if current_sequence_name:
                    sequences[current_sequence_name] = current_sequence

                current_sequence_name = line[1:].strip()
                current_sequence = ""
            else:
                current_sequence += line.strip()

        if current_sequence_name:
            sequences[current_sequence_name] = current_sequence

    return sequences


def download_pdb_multiprocessing(
    pdb_codes: List[str],
    out_dir: Union[str, Path],  # type: ignore
    format: Literal["pdb", "mmtf", "mmcif", "cif", "bcif"] = "pdb",
    overwrite: bool = False,
    strict: bool = False,
    max_workers: int = 16,
    chunksize: int = 32,
) -> List[Path]:
    """Downloads PDB structures in parallel.

    :param pdb_codes: List of PDB codes to download.
    :type pdb_codes: List[str]
    :param out_dir: Path to directory to download PDB structures to.
    :type out_dir: Union[str, Path]
    :param format: Filetype to download. ``pdb``, ``mmtf``, ``mmcif``/``cif`` or ``bcif``.
    :type format: str
    :param overwrite: Whether to overwrite existing files, defaults to
        ``False``.
    :type overwrite: bool
    :param strict: Whether to check for successful download of each file,
        defaults to ``False``.
    :type strict: bool
    :param max_workers: Number of workers to uses, defaults to 16
    :type max_workers: int
    :param chunksize: Chunk to split list into for each worker, defaults to 32
    :type chunksize: int
    :return: List of Paths to downloaded PDB files.
    :rtype: List[Path]
    """
    out_dir: Path = Path(out_dir)
    func = partial(
        download_pdb,
        out_dir=out_dir,
        format=format,
        overwrite=overwrite,
        strict=strict,
    )
    with Pool(processes=max_workers) as pool:
        results = list(
            tqdm(
                pool.imap_unordered(func, pdb_codes, chunksize=chunksize),
                total=len(pdb_codes),
                desc="Downloading PDB files",
                unit="file",
            )
        )
    return results


def download_pdb(
    pdb_code: str,
    out_dir: Optional[Union[str, Path]] = None,
    format: Literal["pdb", "mmtf", "mmcif", "cif", "bcif"] = "pdb",
    check_obsolete: bool = False,
    overwrite: bool = False,
    strict: bool = True,
) -> Path:
    """
    Download PDB structure from PDB.

    If no structure is found, we perform a lookup against the record of
    obsolete PDB codes (https://files.wwpdb.org/pub/pdb/data/status/obsolete.dat)

    :param pdb_code: 4 character PDB accession code.
    :type pdb_code: str
    :param out_dir: Path to directory to download PDB structure to. If ``None``,
        will download to a temporary directory.
    :type out_dir: Optional[Union[str, Path]]
    :param format: Filetype to download. ``pdb``, ``mmtf``, ``mmcif``/``cif`` or ``bcif``.
    :type format: str
    :param check_obsolete: Whether to check for obsolete PDB codes,
        defaults to ``False``. If an obsolete PDB code is found, the updated PDB
        is downloaded.
    :type check_obsolete: bool
    :param overwrite: If ``True``, will overwrite existing files.
    :type overwrite: bool
    :param strict: If ``True``, will raise an exception if the PDB file is not
        found.
    :type strict: bool
    :return: returns filepath to downloaded structure.
    :rtype: Path
    """
    pdb_code = pdb_code.lower()
    if format == "pdb":
        BASE_URL = "https://files.rcsb.org/download/"
        extension = ".pdb"
    elif format == "mmtf":
        BASE_URL = "https://mmtf.rcsb.org/v1.0/full/"
        extension = ".mmtf.gz"
    elif format == "cif" or format == "mmcif":
        BASE_URL = "https://files.rcsb.org/download/"
        extension = ".cif.gz"
    elif format == "bcif":
        BASE_URL = "https://models.rcsb.org/"
        extension = ".bcif.gz"
    else:
        raise ValueError(
            f"Invalid format: {format}. Must be 'pdb', 'mmtf', '(mm)cif' or 'bcif'."
        )

    # Make output directory if it doesn't exist or set it to tempdir if None
    if out_dir is not None:
        out_dir = Path(out_dir)
    else:
        out_dir = Path(tempfile.TemporaryDirectory().name)

    os.makedirs(Path(out_dir), exist_ok=True)

    if check_obsolete:
        obs_map = get_obsolete_mapping()
        try:
            new_pdb = obs_map[pdb_code.lower()].lower()
            log.info(
                f"{pdb_code} is deprecated. Downloading {new_pdb} instead."
            )
            return download_pdb(
                new_pdb, out_dir, format=format, overwrite=overwrite
            )
        except KeyError:
            log.warning(f"PDB {pdb_code} not found. Possibly too large; large \
                    structures are only provided as mmCIF files.")
            return

    # Check if PDB already exists
    if os.path.exists(out_dir / f"{pdb_code}{extension}") and not overwrite:
        log.debug(
            f"{pdb_code} already exists: {out_dir / f'{pdb_code}{extension}'}"
        )
        return out_dir / f"{pdb_code}{extension}"

    # Download
    try:
        wget.download(
            f"{BASE_URL}{pdb_code}{extension}",
            out=str(out_dir / f"{pdb_code}{extension}"),
            bar=None,
        )
    except HTTPError:
        log.warning(f"PDB {pdb_code} not found.")

    # Check file exists
    if strict:
        assert os.path.exists(
            out_dir / f"{pdb_code}{extension}"
        ), f"{pdb_code} download failed. Not found in {out_dir}"
    log.debug(f"{pdb_code} downloaded to {out_dir}")
    return out_dir / f"{pdb_code}{extension}"


def get_protein_name_from_filename(path: str) -> str:
    """
    Extracts a filename from a ``path``

    :param path: Path to extract filename from.
    :type path: str
    :return: file name.
    :rtype: str
    """
    _, tail = os.path.split(path)
    tail = os.path.splitext(tail)[0]
    return tail


def filter_dataframe(
    dataframe: pd.DataFrame,
    by_column: str,
    list_of_values: List[Any],
    boolean: bool,
) -> pd.DataFrame:
    """
    Filter function for DataFrame.

    Filters the DataFrame such that the ``by_column`` values have to be
    in the ``list_of_values`` list if ``boolean == True``, or not in the list
    if ``boolean == False``.

    :param dataframe: pd.DataFrame to filter.
    :type dataframe: pd.DataFrame
    :param by_column: str denoting column of DataFrame to filter.
    :type by_column: str
    :param list_of_values: List of values to filter with.
    :type list_of_values: List[Any]
    :param boolean: indicates whether to keep or exclude matching
        ``list_of_values``. ``True`` -> in list, ``False`` -> not in list.
    :type boolean: bool
    :returns: Filtered DataFrame.
    :rtype: pd.DataFrame
    """
    df = dataframe.copy()
    df = df[df[by_column].isin(list_of_values) == boolean]
    df.reset_index(inplace=True, drop=True)

    return df


def compute_rgroup_dataframe(pdb_df: pd.DataFrame) -> pd.DataFrame:
    """Return the atoms that are in R-groups and not the backbone chain.

    :param pdb_df: DataFrame to compute R group dataframe from.
    :type pdb_df: pd.DataFrame
    :returns: DataFrame containing R-groups only (backbone atoms removed).
    :rtype: pd.DataFrame
    """
    return filter_dataframe(pdb_df, "atom_name", BACKBONE_ATOMS, False)


def download_alphafold_structure(
    uniprot_id: str,
    version: int = 2,
    out_dir: str = ".",
    rename: bool = True,
    pdb: bool = True,
    mmcif: bool = False,
    aligned_score: bool = True,
) -> Union[str, Tuple[str, str], None]:
    """
    Downloads a structure from the Alphafold EBI database
    (https://alphafold.ebi.ac.uk/files/").

    :param uniprot_id: UniProt ID of desired protein.
    :type uniprot_id: str
    :param version: Version of the structure to download
    :type version: int
    :param out_dir: string specifying desired output location. Default is pwd.
    :type out_dir: str
    :param rename: boolean specifying whether to rename the output file to
        ``$uniprot_id.pdb``. Default is ``True``.
    :type rename: bool
    :param pdb: boolean specifying whether to download the PDB file. Default is
        ``True``.
    :type pdb: bool
    :param mmcif: Bool specifying whether to download MMCiF or PDB. Default is
        ``False`` (downloads pdb).
    :type mmcif: bool
    :param retrieve_aligned_score: Bool specifying whether or not to download
        score alignment json.
    :type retrieve_aligned_score: bool
    :return: path to output. Tuple if several outputs specified.
    :rtype: Union[str, Tuple[str, str]]
    """
    BASE_URL = "https://alphafold.ebi.ac.uk/files/"
    uniprot_id = uniprot_id.upper()

    if not mmcif and not pdb:
        raise ValueError("Must specify either mmcif or pdb.")
    if mmcif:
        query_url = f"{BASE_URL}AF-{uniprot_id}-F1-model_v{version}.cif"
    if pdb:
        query_url = f"{BASE_URL}AF-{uniprot_id}-F1-model_v{version}.pdb"

    try:
        structure_filename = wget.download(query_url, out=out_dir)
    except HTTPError:
        log.warning(
            f"No structure found for {uniprot_id}. Used URL: {query_url}"
        )
        return None

    if rename:
        extension = ".pdb" if pdb else ".cif"
        os.rename(
            structure_filename, Path(out_dir) / f"{uniprot_id}{extension}"
        )
        structure_filename = str(
            (Path(out_dir) / f"{uniprot_id}{extension}").resolve()
        )

    log.debug(f"Downloaded AlphaFold PDB file for: {uniprot_id}")
    if aligned_score:
        score_query = (
            BASE_URL
            + "AF-"
            + uniprot_id
            + f"-F1-predicted_aligned_error_v{version}.json"
        )
        score_filename = wget.download(score_query, out=out_dir)
        if rename:
            os.rename(score_filename, Path(out_dir) / f"{uniprot_id}.json")
            score_filename = str(
                (Path(out_dir) / f"{uniprot_id}.json").resolve()
            )
        return structure_filename, score_filename

    return structure_filename


def three_to_one_with_mods(res: str) -> str:
    """
    Converts three letter AA codes into 1 letter. Allows for modified residues.

    See: :const:`~graphein.protein.resi_atoms.RESI_THREE_TO_1`.

    :param res: Three letter residue code string.
    :type res: str
    :return: 1-letter residue code.
    :rtype: str
    """
    return RESI_THREE_TO_1[res]


def save_graph_to_pdb(
    g: nx.Graph,
    path: str,
    gz: bool = False,
    atoms: bool = True,
    hetatms: bool = True,
):
    """Saves processed ``pdb_df`` (``g.graph["pdb_df"]``) dataframe to a PDB
    file.

    N.B. PDBs do not contain connectivity information. This only captures the
    nodes in the graph. Connectivity is filled in according to standard rules by
    visualisation programs.

    :param g: Protein graph to save dataframe from.
    :type g: nx.Graph
    :param path: Path to save PDB file to.
    :type path: str
    :param gz: Whether to gzip the file. Defaults to ``False``.
    :type gz: bool
    """
    ppd = PandasPdb()

    df = g.graph["pdb_df"].copy()
    # format charge correctly
    df.charge = pd.to_numeric(df.charge, errors="coerce")

    # Add blank columns
    blank_cols = [
        "blank_1",
        "blank_2",
        "blank_3",
        "blank_4",
        "segment_id",
    ]
    for col in blank_cols:
        if col not in df.columns:
            df[col] = ""
    df["line_idx"] = list(range(1, len(df) + 1))
    df = df[pdb_df_columns]
    atom_df = filter_dataframe(df, "record_name", ["ATOM"], boolean=True)
    hetatm_df = filter_dataframe(df, "record_name", ["HETATM"], boolean=True)

    if atoms:
        ppd.df["ATOM"] = atom_df
    if hetatms:
        ppd.df["HETATM"] = hetatm_df
    ppd.to_pdb(path=path, records=None, gz=gz, append_newline=True)
    log.debug(f"Successfully saved graph to {path}")


def save_pdb_df_to_pdb(
    df: pd.DataFrame,
    path: str,
    gz: bool = False,
    atoms: bool = True,
    hetatms: bool = True,
):
    """Saves pdb dataframe to a PDB file.

    :param g: DataFrame to save as PDB
    :type g: pd.DataFrame
    :param path: Path to save PDB file to.
    :type path: str
    :param gz: Whether to gzip the file. Defaults to ``False``.
    :type gz: bool
    """
    df = df.copy()
    # format charge correctly
    df.charge = pd.to_numeric(df.charge, errors="coerce")
    df.alt_loc = df.alt_loc.fillna(" ")
    blank_cols = ["blank_1", "blank_2", "blank_3", "blank_4", "segment_id"]
    for col in blank_cols:
        if col not in df.columns:
            df[col] = ""
    df["line_idx"] = list(range(1, len(df) + 1))
    df = df[pdb_df_columns]

    atom_df = filter_dataframe(df, "record_name", ["ATOM"], boolean=True)
    hetatm_df = filter_dataframe(df, "record_name", ["HETATM"], boolean=True)

    ppd = PandasPdb()

    if atoms:
        ppd.df["ATOM"] = atom_df
    if hetatms:
        ppd.df["HETATM"] = hetatm_df
    ppd.to_pdb(path=path, records=None, gz=gz, append_newline=True)
    log.debug(f"Successfully saved PDB dataframe to {path}")


def save_rgroup_df_to_pdb(
    g: nx.Graph,
    path: str,
    gz: bool = False,
    atoms: bool = True,
    hetatms: bool = True,
):
    """Saves R-group (``g.graph["rgroup_df"]``) dataframe to a PDB file.

    N.B. PDBs do not contain connectivity information.
    This only captures the atoms in the r groups.
    Connectivity is filled in according to standard rules by visualisation
    programs.

    :param g: Protein graph to save R group dataframe from.
    :type g: nx.Graph
    :param path: Path to save PDB file to.
    :type path: str
    :param gz: Whether to gzip the file. Defaults to ``False``.
    :type gz: bool
    """
    ppd = PandasPdb()
    df = g.graph["rgroup_df"].copy()

    # format charge correctly
    df.charge = pd.to_numeric(df.charge, errors="coerce")

    blank_cols = ["blank_1", "blank_2", "blank_3", "blank_4", "segment_id"]
    for col in blank_cols:
        if col not in df.columns:
            df[col] = [""] * len(df)
    df["line_idx"] = list(range(1, len(df) + 1))
    df = df[pdb_df_columns]

    atom_df = filter_dataframe(df, "record_name", ["ATOM"], boolean=True)
    hetatm_df = filter_dataframe(df, "record_name", ["HETATM"], boolean=True)

    if atoms:
        ppd.df["ATOM"] = atom_df
    if hetatms:
        ppd.df["HETATM"] = hetatm_df
    ppd.to_pdb(path=path, records=None, gz=gz, append_newline=True)
    log.debug(f"Successfully saved rgroup data to {path}")


def esmfold(
    sequence: str,
    out_path: Optional[str] = None,
    version: int = 1,
    format: str = "pdb",
):
    """Fold a protein sequence using the ESMFold model from the ESMFold server at
    https://api.esmatlas.com/foldSequence/v1/pdb/.


    Parameters
    ----------
    sequence : str
        A protein sequence in one-letter code.
    out_path : str, optional
        Path to save the PDB file to. If `None`, the file is not saved.
        Defaults to `None`.
    version : int, optional
        The version of the ESMFold model to use. Defaults to `1`.
    Returns
    --------
    self
    """
    URL = f"https://api.esmatlas.com/foldSequence/v{version}/{format}/"

    headers: Dict[str, str] = {
        "Content-Type": "application/x-www-form-urlencoded",
    }

    cif = requests.post(URL, data=sequence, headers=headers).text
    # append header
    header = "\n".join(
        [f"data_{sequence}", "#", f"_entry.id\t{sequence}", "#\n"]
    )
    cif = header + cif
    if out_path is not None:
        with open(out_path, "w") as f:
            f.write(cif)


def cast_pdb_column_to_type(
    pdb: PandasPdb, column_name: str, type: Type
) -> PandasPdb:
    """Casts a specified column within a PandasPdb object to a given type
    and returns the typecasted PandasPdb object.

    :param pdb: Input PandasPdb object.
    type pdb: PandasPdb
    :param column_name: Name of column to typecast.
    :type column_name: str
    :param type: Type to which to cast the specified column.
    :type type: Type
    :return: Typecasted PandasPdb object.
    :rtype: PandasPdb
    """
    for key in pdb.df:
        if column_name in pdb.df[key]:
            pdb.df[key][column_name] = pdb.df[key][column_name].apply(type)
    return pdb


def extract_chains_to_file(
    pdb_file: str, chains: List[str], out_dir: str, models: List[int] = [1]
) -> List[str]:
    """Extracts chains from a PDB file to separate files.

    .. code-block::python

        extract_chains_to_file("4hhb.pdb", ["A", "B"], ".", [1])

    This will create new files ``./4hhb_A.pdb`` and ``./4hhb_B.pdb``.


    :param pdb_file: PDB file
    :type pdb_file: str
    :param chains: List of chains to extract
    :type chains: List[str]
    :param out_file: Directory of output files
    :type out_file: str
    :param models: List of indices of models from which to extract chains,
        defaults to ``[1]``.
    :type models: List[int], optional
    :return: List of output file paths
    :rtype: List[str]
    """
    fname = os.path.basename(pdb_file)
    fname = fname.split(".")[0]

    ppdb = PandasPdb().read_pdb(pdb_file).get_models(models)
    # work around int-typing bug for `model_id` within version `0.5.0.dev0` of BioPandas -> appears when calling `to_pdb()`
    cast_pdb_column_to_type(ppdb, column_name="model_id", type=str)

    out_files = []
    for chain in chains:
        out_path = os.path.join(out_dir, f"{fname}_{chain}.pdb")
        out_files.append(out_path)
        df = ppdb.df["ATOM"].loc[ppdb.df["ATOM"]["chain_id"] == chain]
        out_df = PandasPdb()
        out_df.df["ATOM"] = df
        out_df.to_pdb(
            path=out_path, records=None, gz=False, append_newline=True
        )
    return out_files



"""Functions for adding nodes features to a PPI Graph."""

# %%
# Graphein
# Author: Ramon Vinas, Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from typing import Any, Dict

import networkx as nx
from bioservices import HGNC, UniProt
from loguru import logger as log


def add_sequence_to_nodes(n: str, d: Dict[str, Any]):
    """
    Maps UniProt ACC to UniProt ID. Retrieves sequence from UniProt and adds it to the node as a feature

    :param n: Graph node.
    :type n: str
    :param d: Graph attribute dictionary.
    :type d: Dict[str, Any]
    """
    h = HGNC(verbose=False)
    u = UniProt(verbose=False)

    d["uniprot_ids"] = h.fetch("symbol", d["protein_id"])["response"]["docs"][
        0
    ]["uniprot_ids"]

    # Todo these API calls should probably be batched
    # Todo mapping with bioservices to support other protein IDs?

    for id in d["uniprot_ids"]:
        d[f"sequence_{id}"] = u.search(id, columns="sequence").split("\n")[1]


if __name__ == "__main__":
    from functools import partial

    import matplotlib.pyplot as plt

    from graphein.ppi.config import PPIGraphConfig
    from graphein.ppi.edges import add_biogrid_edges, add_string_edges
    from graphein.ppi.graphs import compute_ppi_graph

    protein_list = [
        "CDC42",
        "CDK1",
        "KIF23",
        "PLK1",
        "RAC2",
        "RACGAP1",
        "RHOA",
        "RHOB",
    ]

    config = PPIGraphConfig()
    kwargs = config.kwargs

    g = compute_ppi_graph(
        protein_list=protein_list,
        edge_construction_funcs=[
            partial(add_string_edges, kwargs=kwargs),
            partial(add_biogrid_edges, kwargs=kwargs),
        ],
    )

    for n, d in g.nodes(data=True):
        add_sequence_to_nodes(n, d)

    print(nx.get_node_attributes(g, "sequence"))
    edge_colors = [
        (
            "r"
            if g[u][v]["kind"] == {"string"}
            else "b" if g[u][v]["kind"] == {"biogrid"} else "y"
        )
        for u, v in g.edges()
    ]

    print(nx.info(g))
    nx.draw(g, with_labels=True, edge_color=edge_colors)
    plt.show()



"""Functions for computing graph-level features based on structure."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
import math
from typing import Union

import networkx as nx
import numpy as np

from graphein.protein.resi_atoms import ATOMIC_MASSES


def add_radius_of_gyration(G: nx.Graph, round: bool = False) -> nx.Graph:
    """Adds radius of gyration (Rg) to graph as a graph attribute  (``G.graph["radius_of_gyration"]``).

    Atomic masses are defined in :ref:`graphein.protein.resi_atoms.ATOMIC_MASSES`.

    :param G: Structure graph to add radius of gyration to.
    :type G: nx.Graph
    :param round: Whether to round the result to the nearest integer, defaults to ``False``
    :type round: bool
    :return: Graph with radius of gyration added (in angstroms).
    :rtype: nx.Graph
    """

    G.graph["radius_of_gyration"] = radius_of_gyration(G, round)
    return G


def radius_of_gyration(G: nx.Graph, round: bool = False) -> Union[float, int]:
    """Calculates the radius of gyration of a structure graph in angstroms.

    Atomic masses are defined in :ref:`graphein.protein.resi_atoms.ATOMIC_MASSES`.

    :param G: Graph to calculate radius of gyration of.
    :type G: nx.Graph
    :param round: Whether to round the result to the nearest integer.
    :type round: bool
    :return: Radius of gyration in angstroms.
    :rtype: float
    """
    masses = [
        ATOMIC_MASSES[d["element_symbol"]] for _, d in G.nodes(data=True)
    ]
    total_mass = sum(masses)

    coords = [d["coords"] for _, d in G.nodes(data=True)]
    weighted_coords = [coord * mass for coord, mass in zip(coords, masses)]

    rr = sum(
        np.sum(w_coord * coord)
        for w_coord, coord in zip(weighted_coords, coords)
    )
    mm = sum((sum(w_coord) / total_mass) ** 2 for w_coord in weighted_coords)

    rg = math.sqrt(rr / total_mass - mm)
    return round(rg, 3) if round else rg



"""Functions for adding edges to a PPI Graph from parsed STRING & BIOGRID API call outputs."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Ramon Vinas
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import networkx as nx
import pandas as pd
from loguru import logger as log

from graphein.ppi.parse_biogrid import BIOGRID_df
from graphein.ppi.parse_stringdb import STRING_df


def add_string_edges(G: nx.Graph, **kwargs) -> nx.Graph:
    """
    Adds edges from STRING PPI database (https://string-db.org/) to a PPI Graph.

    :param G: Graph to edges to (populated with ``protein_id`` nodes).
    :type G: nx.Graph
    :param kwargs:  Additional parameters to pass to STRING API calls.
    :return: PPI Graph with STRING interactions added as edges.
    :rtype: nx.Graph
    """
    G.graph["sources"].append("string")
    G.graph["string_df"] = STRING_df(
        G.graph["protein_list"],
        ncbi_taxon_id=G.graph["ncbi_taxon_id"],
        kwargs=kwargs,
    )
    add_interacting_proteins(G, df=G.graph["string_df"], kind="string")

    return G


def add_biogrid_edges(G: nx.Graph, **kwargs) -> nx.Graph:
    """
    Adds edges from the BIOGRID database (https://thebiogrid.org/) to PPI Graph.

    :param G: Graph to edges to (populated with ``protein_id`` nodes).
    :type G: nx.Graph
    :param kwargs:  Additional parameters to pass to BIOGRID API calls.
    :return: nx.Graph PPIGraph with BIOGRID interactions added as edges.
    :rtype: nx.Graph
    """
    G.graph["sources"].append("biogrid")
    G.graph["biogrid_df"] = BIOGRID_df(
        G.graph["protein_list"],
        ncbi_taxon_id=G.graph["ncbi_taxon_id"],
        kwargs=kwargs,
    )
    add_interacting_proteins(G, df=G.graph["biogrid_df"], kind="biogrid")

    return G


def add_interacting_proteins(
    G: nx.Graph, df: pd.DataFrame, kind: str
) -> nx.Graph:
    """
    Generic function for adding interaction edges to PPI Graph.
    You can use this function to additional interactions using a DataFrame with columns ``"p1"`` and ``"p2"``.

    :param G: PPI Graph to populate with edges.
    :type G: nx.Graph
    :param df: DataFrame containing edge list.
    :type df: pd.DataFrame
    :param kind: name of interaction type.
    :type kind: str
    :returns: PPI Graph with pre-computed edges added.
    :rtype: nx.Graph
    """

    protein_1 = df["p1"].values
    protein_2 = df["p2"].values

    interacting_proteins = set(list(zip(protein_1, protein_2)))

    for p1, p2 in interacting_proteins:
        if G.has_edge(p1, p2):
            G.edges[p1, p2]["kind"].add(kind)
        else:
            G.add_edge(p1, p2, kind={kind})
    log.debug(f"Added {len(df)} {kind} interaction edges")

    return G



"""Tests for graphein.protein.tensor.io"""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import os
from pathlib import Path

import pytest
from biopandas.pdb import PandasPdb
from pandas.testing import assert_frame_equal

from graphein.protein.tensor import Protein
from graphein.protein.tensor.io import (
    protein_df_to_chain_tensor,
    protein_df_to_tensor,
    protein_to_pyg,
    to_dataframe,
    to_pdb,
)
from graphein.protein.tensor.sequence import get_residue_id

try:
    import torch

    TORCH_AVAIL = True
except ImportError:
    TORCH_AVAIL = False

PDB_DATA_PATH = (
    Path(__file__).resolve().parent.parent / "test_data" / "4hhb.pdb"
)
CIF_DATA_PATH = (
    Path(__file__).resolve().parent.parent / "test_data" / "4hhb.cif"
)


def get_example_df():
    p = PandasPdb().fetch_pdb("3EIY")
    return p.df["ATOM"]


@pytest.mark.skipif(not TORCH_AVAIL, reason="PyTorch not available")
def test_protein_df_to_chain_tensor():
    df = get_example_df()
    num_chains = len(df.chain_id.unique())

    num_residues = len(get_residue_id(df))

    chains = protein_df_to_chain_tensor(df, one_hot=True)
    print(chains)
    assert (
        chains.shape[0] == num_residues
    ), "Number of residues and chain IDs do not match."
    assert (
        chains.shape[1] == num_chains
    ), "Number of chains do not match dimension"
    assert chains.max() == 1, "Chain IDs are not one hot."
    if num_chains > 1:
        assert chains.min() == 0, "Chain IDs are not one hot."

    chain = protein_df_to_chain_tensor(df, one_hot=False)
    assert (
        chain.shape[0] == num_residues
    ), "Number of residues and chain IDs do not match."
    assert chain.max() == num_chains - 1, "Chain IDs are not zero-indexed."
    assert chain.min() == 0, "Chain IDs are not zero-indexed."


@pytest.mark.skipif(not TORCH_AVAIL, reason="PyTorch not available")
def test_protein_df_to_tensor():  # sourcery skip: extract-duplicate-method
    df = get_example_df()

    num_residues = len(get_residue_id(df))
    print(get_residue_id(df))
    assert num_residues == 174, "Incorrect number of residues."

    positions = protein_df_to_tensor(df)
    assert positions.shape[0] == num_residues, "Incorrect number of residues."
    assert positions.shape[1] == 37, "Incorrect number of atoms."
    assert positions.shape[2] == 3, "Incorrect number of coordinates."

    # Backbone only
    atoms_to_keep = ["N", "CA", "C", "O"]
    positions = protein_df_to_tensor(df, atoms_to_keep=atoms_to_keep)
    assert positions.shape[0] == num_residues, "Incorrect number of residues."
    assert positions.shape[1] == 4, "Incorrect number of atoms."
    assert positions.shape[2] == 3, "Incorrect number of coordinates."


def test_to_pdb():
    protein = Protein().from_pdb_code("4hhb")
    to_pdb(protein.coords, "test.pdb")
    assert os.path.exists("test.pdb"), "File does not exist"

    ppdb1 = PandasPdb().read_pdb("test.pdb")
    ppdb2 = PandasPdb().fetch_pdb("4hhb")

    assert_frame_equal(
        ppdb1.df["ATOM"][["x_coord", "y_coord", "z_coord"]][:50],
        ppdb2.df["ATOM"][["x_coord", "y_coord", "z_coord"]][:50],
    )

    assert_frame_equal(
        ppdb1.df["ATOM"][["atom_name", "residue_name", "element_symbol"]][:50],
        ppdb2.df["ATOM"][["atom_name", "residue_name", "element_symbol"]][:50],
    )


def test_pdb_to_pyg():
    pyg_object = protein_to_pyg(PDB_DATA_PATH)


def test_cif_to_pyg():
    pyg_object = protein_to_pyg(CIF_DATA_PATH)


def test_pdb_and_cif_parsing():
    pdb_pyg = protein_to_pyg(PDB_DATA_PATH)
    cif_pyg = protein_to_pyg(CIF_DATA_PATH)
    assert pdb_pyg.coords.shape == cif_pyg.coords.shape



"""Tests for graphein.protein.tensor.data."""

import pytest

from graphein.protein.tensor.data import Protein

try:
    import torch

    TORCH_AVAIL = True
except ImportError:
    TORCH_AVAIL = False


@pytest.mark.skipif(not TORCH_AVAIL, reason="PyTorch not available")
def test_save_and_load_protein():
    a = Protein().from_pdb_code("4hhb")
    torch.save(a, "4hhb.pt")
    b = torch.load("4hhb.pt", weights_only=False)
    assert a == b



import os

import networkx as nx
from pandas.testing import assert_frame_equal

from graphein.protein.graphs import (
    construct_graph,
    filter_dataframe,
    read_pdb_to_dataframe,
)
from graphein.protein.utils import (
    download_pdb,
    download_pdb_multiprocessing,
    get_obsolete_mapping,
    save_graph_to_pdb,
    save_pdb_df_to_pdb,
    save_rgroup_df_to_pdb,
)


def test_save_graph_to_pdb():
    g = construct_graph(pdb_code="4hhb")

    save_graph_to_pdb(g, "/tmp/test_graph.pdb", hetatms=False)

    a = read_pdb_to_dataframe("/tmp/test_graph.pdb")
    # Check file exists
    assert os.path.isfile("/tmp/test_graph.pdb")

    graph_df = (
        g.graph["pdb_df"]
        .drop(
            [
                "node_id",
                "residue_id",
            ],
            axis=1,
        )
        .reset_index(drop=True)
    )

    a.reset_index(drop=True, inplace=True)
    a = a[graph_df.columns]  # Reorder columns

    assert_frame_equal(
        a,
        graph_df,
    )
    h = construct_graph(path="/tmp/test_graph.pdb")

    # We check for isomorphism rather than equality as array features are not
    # comparable
    assert nx.is_isomorphic(g, h)


def test_save_pdb_df_to_pdb():
    g = construct_graph(pdb_code="4hhb")

    save_pdb_df_to_pdb(g.graph["pdb_df"], "/tmp/test_pdb.pdb", hetatms=False)
    a = read_pdb_to_dataframe("/tmp/test_pdb.pdb")
    # Check file exists
    assert os.path.isfile("/tmp/test_graph.pdb")

    assert_frame_equal(
        a,
        g.graph["pdb_df"]
        .drop(
            [
                "node_id",
                "residue_id",
            ],
            axis=1,
        )
        .reset_index(drop=True),
    )

    # Now check for raw, unprocessed DF
    save_pdb_df_to_pdb(g.graph["raw_pdb_df"], "/tmp/test_pdb.pdb")
    h = construct_graph(path="/tmp/test_pdb.pdb")

    # We check for isomorphism rather than equality as array features are not
    # comparable
    assert nx.is_isomorphic(g, h)


def test_save_rgroup_df_to_pdb():
    g = construct_graph(pdb_code="4hhb")

    save_rgroup_df_to_pdb(g, "/tmp/test_rgroup.pdb", hetatms=False)
    a = read_pdb_to_dataframe("/tmp/test_rgroup.pdb")
    # Check file exists
    assert os.path.isfile("/tmp/test_rgroup.pdb")

    # We drop the line_idx columns as these will be renumbered
    assert_frame_equal(
        a,
        filter_dataframe(
            g.graph["rgroup_df"], "record_name", ["HETATM"], False
        ).drop(["node_id", "residue_id"], axis=1),
    )


def test_download_obsolete_structure():
    fp = download_pdb(pdb_code="116L", check_obsolete=True)
    assert os.path.exists(fp)
    assert str(fp).endswith("216l.pdb")


def test_download_structure():
    fp = download_pdb(pdb_code="4hhb")
    assert os.path.exists(fp)
    assert str(fp).endswith("4hhb.pdb")


def test_download_structure_multi():
    fps = download_pdb_multiprocessing(pdb_codes=["4hhb", "4hhb"], out_dir=".")
    for path in fps:
        assert os.path.exists(path)
        assert str(path).endswith("4hhb.pdb")


def test_download_obsolete_map():
    mapping = get_obsolete_mapping()
    assert len(mapping) > 100


if __name__ == "__main__":
    test_save_graph_to_pdb()
    test_save_pdb_df_to_pdb()
    test_save_rgroup_df_to_pdb()



from pathlib import Path

import networkx as nx
import pytest
from hypothesis import given
from hypothesis.strategies import text

from graphein.rna.edges import (
    add_all_dotbracket_edges,
    add_base_pairing_interactions,
    add_phosphodiester_bonds,
)
from graphein.rna.graphs import (
    RNA_BASES,
    SUPPORTED_DOTBRACKET_NOTATION,
    construct_graph,
)

TEST_SEQUENCE = "UUGGAGUACACAACCUGUACACUCUUUC"
TEST_DOTBRACKET = "..(((((..(((...)))..)))))..."


def test_construct_rna_graph():
    g = construct_graph(
        dotbracket=TEST_DOTBRACKET,
        sequence=TEST_SEQUENCE,
        edge_construction_funcs=[
            add_base_pairing_interactions,
            add_phosphodiester_bonds,
        ],
    )

    h = construct_graph(
        dotbracket=TEST_DOTBRACKET,
        sequence=TEST_SEQUENCE,
        edge_construction_funcs=[add_all_dotbracket_edges],
    )

    assert g.edges() == h.edges()
    assert g.nodes() == h.nodes()

    # Check number of nodes and edges
    assert len(g.nodes()) == len(TEST_SEQUENCE)
    assert len(g.nodes()) == len(TEST_DOTBRACKET)

    # Check node features are in alphabet
    for n, d in g.nodes(data=True):
        assert d["dotbracket_symbol"] in SUPPORTED_DOTBRACKET_NOTATION
        assert d["nucleotide"] in RNA_BASES

    # Count edges for each type
    phosphodiesters = 0
    bp = 0
    for u, v, d in g.edges(data=True):
        if d["attr"] == "phosphodiester_bond":
            phosphodiesters += 1
        elif d["attr"] == "base_pairing":
            bp += 1

    assert phosphodiesters == len(TEST_SEQUENCE) - 1
    assert bp == TEST_DOTBRACKET.count("(")


def test_pdb_rna_graph():
    g = construct_graph(pdb_code="2jyf")
    assert isinstance(g, nx.Graph)

    for n, d in g.nodes(data=True):
        assert isinstance(n, str)
        assert "coords" in d.keys()
        assert "chain_id" in d.keys()
        assert "residue_name" in d.keys()
        assert "residue_number" in d.keys()
        assert "atom_type" in d.keys()
        assert "b_factor" in d.keys()

    for _, _, d in g.edges(data=True):
        assert "kind" in d.keys()
        assert "bond_length" in d.keys()

    assert "name" in g.graph.keys()
    assert "pdb_code" in g.graph.keys()
    assert "path" in g.graph.keys()
    assert "chain_ids" in g.graph.keys()
    assert "pdb_df" in g.graph.keys()
    assert "raw_pdb_df" in g.graph.keys()
    assert "coords" in g.graph.keys()


def test_construct_graph():
    """Example-based test that graph construction works correctly.
    Uses 4hhb PDB file as an example test case.
    """
    file_path = Path(__file__).parent / "test_data/2jyf.pdb"
    g = construct_graph(path=str(file_path))

    for n, d in g.nodes(data=True):
        assert isinstance(n, str)
        assert "coords" in d.keys()
        assert "chain_id" in d.keys()
        assert "residue_name" in d.keys()
        assert "residue_number" in d.keys()
        assert "atom_type" in d.keys()
        assert "b_factor" in d.keys()

    for _, _, d in g.edges(data=True):
        assert "kind" in d.keys()
        assert "bond_length" in d.keys()

    assert "name" in g.graph.keys()
    assert "pdb_code" in g.graph.keys()
    assert "path" in g.graph.keys()
    assert "chain_ids" in g.graph.keys()
    assert "pdb_df" in g.graph.keys()
    assert "raw_pdb_df" in g.graph.keys()
    assert "coords" in g.graph.keys()



"""Tests for graphein.cli"""

import pickle
import tempfile
from pathlib import Path

import networkx as nx
from click.testing import CliRunner

from graphein.cli import main

DATA_PATH = Path(__file__).resolve().parent


def test_cli():
    """Example-based test that graph construction works correctly.

    Uses 4hhb PDB file as an example test case.
    """
    file_path = DATA_PATH.parent / "protein/test_data/4hhb.pdb"
    temp_dir = tempfile.TemporaryDirectory()

    runner = CliRunner()
    result = runner.invoke(main, ["-p", file_path, "-o", Path(temp_dir.name)])
    assert result.exit_code == 0

    with open(Path(temp_dir.name) / f"{file_path.stem}.gpickle", "rb") as f:
        G = pickle.load(f)
    assert isinstance(G, nx.Graph)
    assert len(G) == 574
    # Check number of peptide bonds
    peptide_bond_edges = [
        (u, v)
        for u, v, d in G.edges(data=True)
        if d["kind"] == {"peptide_bond"}
    ]
    assert len(peptide_bond_edges) == 570
    temp_dir.cleanup()



"""Provides geometry-based featurisation functions."""

# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

from typing import List

import networkx as nx
import numpy as np
from loguru import logger as log

from graphein.protein.utils import compute_rgroup_dataframe, filter_dataframe

VECTOR_FEATURE_NAMES: List[str] = [
    "sidechain_vector",
    "c_beta_vector",
    "sequence_neighbour_vector_n_to_c",
    "sequence_neighbour_vector_c_to_n",
    "virtual_c_beta_vector",
]
"""Names of all vector features from the module."""


def add_sidechain_vector(
    g: nx.Graph, scale: bool = True, reverse: bool = False
):
    """Adds vector from node to average position of sidechain atoms.

    We compute the mean of the sidechain atoms for each node. For this we use
    the ``rgroup_df`` dataframe. If the graph does not contain the ``rgroup_df``
    dataframe, we compute it from the ``raw_pdb_df``. If ``scale``, we scale
    the vector to the unit vector. If ``reverse`` is ``True``, we reverse the
    vector (``sidechain - node``). If reverse is false (default) we compute
    (``node - sidechain``).

    :param g: Graph to add vector to.
    :type g: nx.Graph
    :param scale: Scale vector to unit vector. Defaults to ``True``.
    :type scale: bool
    :param reverse: Reverse vector. Defaults to ``False``.
    :type reverse: bool
    """
    # Get or compute R-Group DF
    if "rgroup_df" not in g.graph.keys():
        g.graph["rgroup_df"] = compute_rgroup_dataframe(g.graph["raw_pdb_df"])

    sc_centroid = (
        g.graph["rgroup_df"].groupby("node_id").mean(numeric_only=True)
    )

    # Iterate over nodes and compute vector
    for n, d in g.nodes(data=True):
        if d["residue_name"] == "GLY":
            # If GLY, set vector to 0
            vec = np.array([0.0, 0.0, 0.0])
        elif n not in sc_centroid.index:
            vec = np.array([0.0, 0.0, 0.0])
            log.warning(
                f"Non-glycine residue {n} does not have side-chain atoms."
            )
        else:
            if reverse:
                vec = d["coords"] - np.array(
                    sc_centroid.loc[n][["x_coord", "y_coord", "z_coord"]],
                    dtype=float,
                )
            else:
                vec = (
                    np.array(
                        sc_centroid.loc[n][["x_coord", "y_coord", "z_coord"]],
                        dtype=float,
                    )
                    - d["coords"]
                )

            if scale:
                vec = vec / np.linalg.norm(vec)

        d["sidechain_vector"] = vec


def add_beta_carbon_vector(
    g: nx.Graph, scale: bool = True, reverse: bool = False
):
    """Adds vector from node (typically alpha carbon) to position of beta
    carbon.

    Glycine does not have a beta carbon, so we set it to
    ``np.array([0., 0., 0.])``. We extract the position of the beta carbon from the
    unprocessed atomic PDB dataframe. For this we use the ``raw_pdb_df``
    DataFrame. If ``scale``, we scale the vector to the unit vector. If
    ``reverse`` is ``True``, we reverse the vector (``C beta - node``).
    If ``reverse`` is ``False`` (default) we compute (``node - C beta``).

    :param g: Graph to add vector to.
    :type g: nx.Graph
    :param scale: Scale vector to unit vector. Defaults to ``True``.
    :type scale: bool
    :param reverse: Reverse vector. Defaults to ``False``.
    :type reverse: bool
    """
    # Get or compute R-Group DF
    if "rgroup_df" not in g.graph.keys():
        g.graph["rgroup_df"] = compute_rgroup_dataframe(g.graph["raw_pdb_df"])

    c_beta_coords = filter_dataframe(
        g.graph["rgroup_df"], "atom_name", ["CB"], boolean=True
    )
    c_beta_coords.index = c_beta_coords["node_id"]

    # Iterate over nodes and compute vector
    for n, d in g.nodes(data=True):
        if d["residue_name"] == "GLY":
            vec = np.array([0.0, 0.0, 0.0])
        elif n not in c_beta_coords.index:
            vec = np.array([0.0, 0.0, 0.0])
            log.warning(
                f"Non-glycine residue {n} does not have a beta-carbon."
            )
        else:
            if reverse:
                vec = d["coords"] - np.array(
                    c_beta_coords.loc[n][["x_coord", "y_coord", "z_coord"]],
                    dtype=float,
                )
            else:
                vec = (
                    np.array(
                        c_beta_coords.loc[n][
                            ["x_coord", "y_coord", "z_coord"]
                        ],
                        dtype=float,
                    )
                    - d["coords"]
                )

            if scale:
                vec = vec / np.linalg.norm(vec)
        d["c_beta_vector"] = vec


def add_sequence_neighbour_vector(
    g: nx.Graph, scale: bool = True, reverse: bool = False, n_to_c: bool = True
):
    """Computes vector from node to adjacent node in sequence.
    Typically used with ``CA`` (alpha carbon) graphs.

    If ``n_to_c`` is ``True`` (default), we compute the vectors from the N
    terminus to the C terminus (canonical direction). If ``reverse`` is
    ``False`` (default), we compute ``Node_i - Node_{i+1}``. If ``reverse is
    ``True``, we compute ``Node_{i+1} - Node_i``.

    :param g: Graph to add vector to.
    :type g: nx.Graph
    :param scale: Scale vector to unit vector. Defaults to ``True``.
    :type scale: bool
    :param reverse: Reverse vector. Defaults to ``False``.
    :type reverse: bool
    :param n_to_c: Compute vector from N to C or C to N. Defaults to ``True``.
    :type n_to_c: bool
    """
    suffix = "n_to_c" if n_to_c else "c_to_n"
    # Iterate over every chain
    for chain_id in g.graph["chain_ids"]:
        # Find chain residues
        chain_residues = [
            (n, v) for n, v in g.nodes(data=True) if v["chain_id"] == chain_id
        ]

        if not n_to_c:
            chain_residues.reverse()

        # Iterate over every residue in chain
        for i, residue in enumerate(chain_residues):
            # Checks not at chain terminus - is this versatile enough?
            if i == len(chain_residues) - 1:
                residue[1][f"sequence_neighbour_vector_{suffix}"] = np.array(
                    [0.0, 0.0, 0.0]
                )
                continue

            # Get insertion codes
            ins_current = (
                residue[0].split(":")[3] if residue[0].count(":") > 2 else ""
            )
            ins_next = (
                chain_residues[i + 1][0].split(":")[3]
                if chain_residues[i + 1][0].count(":") > 2
                else ""
            )
            if not n_to_c:
                ins_current, ins_next = ins_next, ins_current

            # Get sequence distance
            dist = abs(
                residue[1]["residue_number"]
                - chain_residues[i + 1][1]["residue_number"]
            )

            # Asserts residues are adjacent
            cond_adjacent = (
                dist == 1
                or (dist == 0 and not ins_current and ins_next == "A")
                or (
                    dist == 0
                    and ins_current
                    and ins_next
                    and chr(ord(ins_current) + 1) == ins_next
                )
            )

            # If this checks out, we compute the non-zero vector
            if cond_adjacent:
                vec = chain_residues[i + 1][1]["coords"] - residue[1]["coords"]

                if reverse:
                    vec = -vec
                if scale:
                    vec = vec / np.linalg.norm(vec)
            else:
                vec = np.array([0.0, 0.0, 0.0])

            residue[1][f"sequence_neighbour_vector_{suffix}"] = vec


def add_virtual_beta_carbon_vector(
    g: nx.Graph, scale: bool = False, reverse: bool = False
):
    """For each node adds a vector from alpha carbon to virtual beta carbon.
    :param g: Graph to add vector to.
    :type g: nx.Graph
    :param scale: Scale vector to unit vector. Defaults to ``False``.
    :type scale: bool
    :param reverse: Reverse vector. Defaults to ``False``.
    :type reverse: bool
    """
    # Get coords of backbone atoms
    coord_dfs = {}
    for atom_type in ["N", "CA", "C"]:
        df = filter_dataframe(
            g.graph["raw_pdb_df"], "atom_name", [atom_type], boolean=True
        )
        df.index = df["node_id"]
        coord_dfs[atom_type] = df

    # Iterate over nodes and compute vector
    for n, d in g.nodes(data=True):
        if any([n not in df.index for df in coord_dfs.values()]):
            vec = np.array([0, 0, 0], dtype=float)
            log.warning(f"Missing backbone atom in residue {n}.")
        else:
            N = np.array(
                coord_dfs["N"].loc[n][["x_coord", "y_coord", "z_coord"]],
                dtype=float,
            )
            Ca = np.array(
                coord_dfs["CA"].loc[n][["x_coord", "y_coord", "z_coord"]],
                dtype=float,
            )
            C = np.array(
                coord_dfs["C"].loc[n][["x_coord", "y_coord", "z_coord"]],
                dtype=float,
            )
            b = Ca - N
            c = C - Ca
            a = np.cross(b, c)
            Cb = -0.58273431 * a + 0.56802827 * b - 0.54067466 * c + Ca
            vec = Cb - Ca

            if reverse:
                vec = -vec
            if scale:
                vec = vec / np.linalg.norm(vec)
        d["virtual_c_beta_vector"] = vec



"""Functions for graph-level featurization of the sequence of a protein. This submodule is focussed on physicochemical
proporties of the sequence."""

# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from functools import partial
from typing import List, Optional

import networkx as nx
import numpy as np
from Bio import SeqUtils
from multipledispatch import dispatch

from graphein.protein.features.sequence.utils import (
    aggregate_feature_over_chains,
    compute_feature_over_chains,
)


@dispatch(str, str)
def molecular_weight(protein: str, seq_type: str = "protein"):
    func = partial(SeqUtils.molecular_weight, seq_type=seq_type)

    return func(protein)


@dispatch(nx.Graph, seq_type=str)
def molecular_weight(protein, seq_type: str = "protein"):
    func = partial(SeqUtils.molecular_weight, seq_type=seq_type)

    return compute_feature_over_chains(
        protein, func, feature_name="molecular_weight"
    )


def get_sinusoid_encoding_table(
    n_position: int, d_hid: int, padding_idx: Optional[int] = None
) -> np.ndarray:
    """
    Numpy-based implementation of sinusoid position encoding used in Transformer models.

    Based on implementation by @foowaa (https://gist.github.com/foowaa/5b20aebd1dff19ee024b6c72e14347bb)

    :param n_position: Number of positions to encode (length of graph) (``N``).
    :param d_hid: dimension of embedding vector (``M``).
    :param padding_idx: Set 0 dimension. Defaults to ``None``.
    :return: Sinusoid table. (``NxM``).
    :rtype: np.ndarray
    """

    def calc_angle(position: int, hid_idx: int) -> float:
        return position / np.power(10000, 2 * (hid_idx // 2) / d_hid)

    def get_pos_angle_vec(position: int) -> List[float]:
        return [calc_angle(position, hid_j) for hid_j in range(d_hid)]

    sinusoid_table = np.array(
        [get_pos_angle_vec(pos_i) for pos_i in range(n_position)]
    )

    sinusoid_table[:, 0::2] = np.sin(sinusoid_table[:, 0::2])  # dim 2i
    sinusoid_table[:, 1::2] = np.cos(sinusoid_table[:, 1::2])  # dim 2i+1

    if padding_idx is not None:
        # zero vector for padding dimension
        sinusoid_table[padding_idx] = 0.0

    return sinusoid_table


def add_positional_encoding(
    G: nx.Graph,
    d_hid: int,
    padding_idx: Optional[int] = None,
    add_to_nodes: bool = True,
) -> nx.Graph:
    """Adds transformer positional encoding (based on sequence) to graph.

    Accessed via: ``g.graph["positional_encoding"]`` if added to the graph
    (``add_to_nodes=False``) or ``d["positional_encoding"] for _, d in g.nodes(data=True)``
    if added to the nodes (``add_to_nodes=True``).

    Nodes are numbered as they occur in list(g.nodes()). By default,
    this corresponds to stacked N->C sequences (ie. in multichain graphs:
    ``SeqA N->C, SeqB N->C, SeqC N->C``).

    :param G: Graph to add positional encoding to.
    :type G: nx.Graph
    :param d_hid: Dimensionality of positional encoding.
    :type d_hid: int
    :param padding_idx: Set 0 dimension. Defaults to ``None``.
    :type padding_idx: Optional[int], optional
    :param add_to_nodes: Whether to add the positional encoding to the graph as a graph-level feature
        (``Nxd_hid`` matrix, where ``N`` is the length (number of nodes) of the graph), or as sliced arrays
        (size ``d_hid``) added to the nodes, defaults to ``True``.
    :type add_to_nodes: bool
    :return: Graph with positional encoding added as either a graph feature or node features.
    :rtype: nx.Graph
    """
    n = len(G)
    sinusoid_table = get_sinusoid_encoding_table(n, d_hid, padding_idx)
    if not add_to_nodes:
        G.graph["positional_encoding"] = sinusoid_table
    else:
        for i, (n, d) in enumerate(G.nodes(data=True)):
            d["positional_encoding"] = sinusoid_table[i, :]

    return G



"""Yaml parser for config objects"""

from functools import partial
from pathlib import Path
from typing import Callable, Union

import yaml
from pydantic import BaseModel

from graphein.grn.config import GRNGraphConfig, RegNetworkConfig, TRRUSTConfig
from graphein.molecule.config import MoleculeGraphConfig
from graphein.ppi.config import BioGridConfig, PPIGraphConfig, STRINGConfig
from graphein.protein.config import (
    DSSPConfig,
    GetContactsConfig,
    ProteinGraphConfig,
    ProteinMeshConfig,
)


def config_constructor(
    loader: yaml.FullLoader, node: yaml.nodes.MappingNode
) -> BaseModel:
    """Construct a BaseModel config.

    :param loader: Given yaml loader
    :param type: yaml.FullLoader
    :param node: A mapping node
    :param type: yaml.nodes.MappingNode
    """
    arg_map = loader.construct_mapping(node, deep=True) if node.value else {}
    return eval(node.tag[1:])(**arg_map)


def function_constructor(
    loader: yaml.FullLoader,
    tag_suffix: str,
    node: Union[yaml.nodes.MappingNode, yaml.nodes.ScalarNode],
) -> Callable:
    """Construct a Callable. If function parameters are given, this returns a partial function.

    :param loader: Given yaml loader
    :param type: yaml.FullLoader
    :param tag_suffix: The name after the !func: tag
    :param type: str
    :param node: A mapping node if function parameters are given, a scalar node if not
    :param type: Union[yaml.nodes.MappingNode, yaml.nodes.ScalarNode]
    """
    arg_map = None
    if isinstance(node, yaml.nodes.MappingNode):
        arg_map = (
            loader.construct_mapping(node, deep=True) if node.value else {}
        )
        node = yaml.nodes.ScalarNode(
            node.tag, "", node.start_mark, node.end_mark
        )
    func = loader.construct_python_name(tag_suffix, node)
    if arg_map:
        func = partial(func, **arg_map)
    return func


def get_loader() -> yaml.Loader:
    """Add constructors to PyYAML loader."""
    # For Python-specific tags, you can use full_load(), which resolves all tags except those known to be unsafe;
    # this includes all the tags listed here: https://pyyaml.org/wiki/PyYAMLDocumentation#yaml-tags-and-python-types
    loader = yaml.FullLoader
    configs = [
        ProteinGraphConfig.__name__,
        DSSPConfig.__name__,
        ProteinMeshConfig.__name__,
        GetContactsConfig.__name__,
        TRRUSTConfig.__name__,
        RegNetworkConfig.__name__,
        GRNGraphConfig.__name__,
        STRINGConfig.__name__,
        BioGridConfig.__name__,
        PPIGraphConfig.__name__,
        MoleculeGraphConfig.__name__,
    ]
    for config in configs:
        loader.add_constructor(f"!{config}", config_constructor)
    loader.add_multi_constructor("!func:", function_constructor)
    return loader


def parse_config(path: Path) -> BaseModel:
    """
    Parses a yaml configuration file into a config object.

    :param path: Path to configuration file
    :type path: pathlib.Path
    """
    with open(path, "rb") as f:
        yml_config = yaml.load(f, Loader=get_loader())
    return yml_config
